#ifndef MAINWINDOW_H_INCLUDED
#define MAINWINDOW_H_INCLUDED

#include <QMainWindow>

#include "Common.h"
#include "Emulator.h"
#include "UKNCBTL.h"
#include "ScreenView.h"
#include "KeyboardView.h"
#include "ui_MainWindow.h"
#include "ui_AboutDlg.h"

class CentralWidget: public QWidget {
public:
  CentralWidget( QWidget *parent = 0 ): QWidget(parent) {
    screenView = new ScreenView( this );
    screenView->move( 0, 0 );

    keyboardView = new KeyboardView( screenView, this );

    adjustLayout();
    show();
  }

  void adjustLayout() {
    keyboardView->move( (screenView->width() - keyboardView->width()) / 2,
	screenView->height() + 8 );

    resize( childrenRect().size() );
  }

  void systemFrame() {
	screenView->prepareScreen();
	screenView->update();
  }

  ScreenView *screenView;
  KeyboardView *keyboardView;
};

class AboutDlg: public QDialog, public Ui_AboutDlg {
public:
  AboutDlg( QWidget *parent = 0 ): QDialog( parent ) {
	setupUi( this );
  }
};

class MainWindow: public QMainWindow, public Ui_MainWindow {
  Q_OBJECT
public:
  MainWindow();
  void restoreSettings();
  void updateMenu();
  ScreenView *screenView() { return cw->screenView; }
  void systemFrame() { cw->systemFrame(); }

protected:
  void adjustLayout();

  CentralWidget *cw;
  void showHideKeyboard( bool show ) {
	cw->keyboardView->setVisible( show );
	adjustLayout();
  }

  void doViewScreenMode( ScreenViewMode mode ) {
	Settings_SetScreenViewMode( mode );
	screenView()->setMode( mode );
  }
  void doViewHeightMode( int mode ) {
	Settings_SetScreenHeightMode( mode );
	screenView()->setHeightMode( mode );
	adjustLayout();
  }

  void doEmulatorFloppy( int slot );
  void doEmulatorCartridge( int slot );

protected slots:
  void doFileLoadState();
  void doFileSaveState();
  void doFileScreenshot();
  void doViewKeyboard( bool show ) {
	Settings_SetKeyboard( show );
	showHideKeyboard( show );
  }
  void doViewScreenRGB() { doViewScreenMode( RGBScreen ); }
  void doViewScreenGRB() { doViewScreenMode( GRBScreen ); }
  void doViewScreenGray() { doViewScreenMode( GrayScreen ); }
  void doViewHeightNormal() { doViewHeightMode( 1 ); }
  void doViewHeightDouble() { doViewHeightMode( 2 ); }
  void doEmulatorRun() {
	if ( g_okEmulatorRunning ) Emulator_Stop();
	else Emulator_Start();
  }
  void doEmulatorAutostart( bool on ) {
	Settings_SetAutostart( on );
  }
  void doEmulatorReset() { Emulator_Reset(); }
  void doEmulatorFloppy0() { doEmulatorFloppy( 0 ); }
  void doEmulatorFloppy1() { doEmulatorFloppy( 1 ); }
  void doEmulatorFloppy2() { doEmulatorFloppy( 2 ); }
  void doEmulatorFloppy3() { doEmulatorFloppy( 3 ); }
  void doEmulatorCartridge1() { doEmulatorCartridge( 1 ); }
  void doEmulatorCartridge2() { doEmulatorCartridge( 2 ); }
  void showAboutDlg() {
	AboutDlg d;
        d.exec();
  }
};

#endif // MAINWINDOW_H_INCLUDED
