#include <QPainter>
#include <QKeyEvent>
#include "ScreenView.h"
#include "Emulator.h"

// Table for color conversion yrgb (4 bits) -> DWORD (32 bits)
const DWORD ScreenView::StandardRGBColors[16] = {
    0x000000, 0x000080, 0x008000, 0x008080,
    0x800000, 0x800080, 0x808000, 0x808080,
    0x000000, 0x0000FF, 0x00FF00, 0x00FFFF,
    0xFF0000, 0xFF00FF, 0xFFFF00, 0xFFFFFF,
};

const DWORD ScreenView::StandardGRBColors[16] = {
    0x000000, 0x000080, 0x800000, 0x800080,
    0x008000, 0x008080, 0x808000, 0x808080,
    0x000000, 0x0000FF, 0xFF0000, 0xFF00FF,
    0x00FF00, 0x00FFFF, 0xFFFF00, 0xFFFFFF,
};

// Table for color conversion, gray (black and white) display
const DWORD ScreenView::GrayColors[16] = {
    0x000000, 0x242424, 0x484848, 0x6C6C6C,
    0x909090, 0xB4B4B4, 0xD8D8D8, 0xFFFFFF,
    0x000000, 0x242424, 0x484848, 0x6C6C6C,
    0x909090, 0xB4B4B4, 0xD8D8D8, 0xFFFFFF,
};

ScreenView::ScreenView( QWidget *parent ):
	QWidget( parent ),
	img( UKNC_SCREEN_WIDTH, UKNC_SCREEN_HEIGHT, QImage::Format_RGB32 ) {
  m_ScreenMode = RGBScreen;
  m_ScreenHeightMode = 1;
  img.fill( 0 );

  m_ScreenKeyQueueTop    = 0;
  m_ScreenKeyQueueBottom = 0;
  m_ScreenKeyQueueCount  = 0;

  resize( img.size() );
  show();
}

void ScreenView::paintEvent( QPaintEvent * /*ev*/ ) {
  QPainter p( this );
  if ( m_ScreenHeightMode == 1 )
	p.drawImage( 0, 0, img.mirrored() );
  else
	p.drawImage( 0, 0, img.mirrored().scaled( size() ) );
}

void ScreenView::prepareScreen() {
  const DWORD* colors;
  switch (m_ScreenMode) {
	case RGBScreen:   colors = StandardRGBColors; break;
	case GrayScreen:  colors = GrayColors; break;
	case GRBScreen:   colors = StandardGRBColors; break;
	default:          colors = StandardRGBColors; break;
  }

  BYTE cursorYRGB = 0;
  BOOL okCursorType = FALSE;
  BYTE cursorPos = 128;
  BOOL cursorOn = FALSE;
  BYTE cursorAddress = 0;  // Address of graphical cursor
  WORD address = 0000270;  // Tag sequence start address
  BOOL okTagSize = FALSE;  // Tag size: TRUE - 4-word, FALSE - 2-word (first tag is always 2-word)
  BOOL okTagType = FALSE;  // Type of 4-word tag: TRUE - set palette, FALSE - set params
  int scale = 1;           // Horizontal scale: 1, 2, 4, or 8
  DWORD palette = 0;       // Palette
  BYTE pbpgpr = 7;         // 3-bit Y-value modifier

  for (int yy = 0; yy < 307; yy++) {
	if (okTagSize) {  // 4-word tag
            WORD tag1 = g_pBoard->GetRAMWord(0, address);
            address += 2;
            WORD tag2 = g_pBoard->GetRAMWord(0, address);
            address += 2;

            if (okTagType)  // 4-word palette tag
                palette = tag1 | (tag2 << 16);

            else { // 4-word params tag
                scale = (tag2 >> 4) & 3;  // Bits 4-5 - new scale value
                //TODO: use Y-value modifier
                pbpgpr = tag2 & 7;  // Y-value modifier
                cursorYRGB = tag1 & 15;  // Cursor color
                okCursorType = ((tag1 & 16) != 0);  // TRUE - graphical cursor, FALSE - symbolic cursor
                cursorPos = ((tag1 >> 8) >> scale) & 0x7f;  // Cursor position in the line
                //TODO: Use cursorAddress
                cursorAddress = (tag1 >> 5) & 7;
                scale = 1 << scale;

            }
        }

        WORD addressBits = g_pBoard->GetRAMWord(0, address);  // The word before the last word - is address of bits from all three memory planes
        address += 2;

        // Calculate size, type and address of the next tag
        WORD tagB = g_pBoard->GetRAMWord(0, address);  // Last word of the tag - is address and type of the next tag
        okTagSize = (tagB & 2) != 0;  // Bit 1 shows size of the next tag
        if (okTagSize) {
            address = tagB & ~7;
            okTagType = (tagB & 4) != 0;  // Bit 2 shows type of the next tag
        }
        else
            address = tagB & ~3;

        if ((tagB & 1) != 0)
            cursorOn = !cursorOn;

        // Draw bits into the bitmap, from line 20 to line 307
        if (yy >= 19 && yy <= 306) {
            // Loop thru bits from addressBits, planes 0,1,2
            // For each pixel:
            //   Get bit from planes 0,1,2 and make value
            //   Map value to palette; result is 4-bit value YRGB
            //   Translate value to 24-bit RGB
            //   Put value to m_bits; repeat using scale value

            int x = 0;
            int y = yy - 19;
            DWORD* pBits = (DWORD*)img.bits() + (288 - 1 - y) * 640;

            for (int pos = 0; ; pos++) {
                // Get bit from planes 0,1,2
                BYTE src0 = g_pBoard->GetRAMByte(0, addressBits);
                BYTE src1 = g_pBoard->GetRAMByte(1, addressBits);
                BYTE src2 = g_pBoard->GetRAMByte(2, addressBits);

                // Loop through the bits of the byte
                for (int bit = 0; bit < 8; bit++) {
                    // Make 3-bit value from the bits
                    BYTE value012 = (src0 & 1) | (src1 & 1) * 2 | (src2 & 1) * 4;
                    // Map value to palette; result is 4-bit value YRGB
                    BYTE valueYRGB;
                    if (cursorOn && pos == cursorPos && (!okCursorType || (okCursorType && bit == cursorAddress)))
                        valueYRGB = cursorYRGB;
                    else
                        valueYRGB = (BYTE) (palette >> (value012 * 4)) & 15;
                    DWORD valueRGB = colors[valueYRGB];

                    // Put value to m_bits; repeat using scale value
                    for (int s = 0; s < scale; s++)
                        *pBits++ = valueRGB;
                    x += scale;

                    // Shift to the next bit
                    src0 = src0 >> 1;
                    src1 = src1 >> 1;
                    src2 = src2 >> 1;
                }

                if (x >= 640) break;  // End of line
                addressBits++;  // Go to the next byte
            }
        }
    }
}

void ScreenView::setHeightMode( int newMode ) {

  m_ScreenHeightMode = newMode;

  if ( m_ScreenHeightMode == 1 )
	resize( img.size() );
  else
	resize( img.width(), img.height() * 2 );
}

void ScreenView::putKeyEventToQueue(BYTE keycode, bool pressed) {
  putKeyEventToQueue( keycode | (pressed? 128 << 8: 0) );
}

void ScreenView::putKeyEventToQueue(WORD keyevent)
{
    if (m_ScreenKeyQueueCount == KEYEVENT_QUEUE_SIZE) return;  // Full queue

    m_ScreenKeyQueue[m_ScreenKeyQueueTop] = keyevent;
    m_ScreenKeyQueueTop++;
    if (m_ScreenKeyQueueTop >= KEYEVENT_QUEUE_SIZE)
        m_ScreenKeyQueueTop = 0;
    m_ScreenKeyQueueCount++;
}

WORD ScreenView::getKeyEventFromQueue() 
{
    if (m_ScreenKeyQueueCount == 0) return 0;  // Empty queue
 
    WORD keyevent = m_ScreenKeyQueue[m_ScreenKeyQueueBottom];
    m_ScreenKeyQueueBottom++;
    if (m_ScreenKeyQueueBottom >= KEYEVENT_QUEUE_SIZE)
        m_ScreenKeyQueueBottom = 0;
    m_ScreenKeyQueueCount--;
 
    return keyevent; 
}

BYTE ScreenView::translateKeyToUkncScan( int key, int modifiers ) {
  if ( modifiers & Qt::KeypadModifier ) {
	switch ( key ) {
	case Qt::Key_1: return 0127;
	case Qt::Key_2: return 0147;
	case Qt::Key_3: return 0167;
	case Qt::Key_4: return 0130;
	case Qt::Key_5: return 0150;
	case Qt::Key_6: return 0170;
	case Qt::Key_7: return 0125;
	case Qt::Key_8: return 0145;
	case Qt::Key_9: return 0165;
	case Qt::Key_0: return 0126;
	case Qt::Key_Plus:   return 0131;
	case Qt::Key_Period: return 0110;
        case Qt::Key_Enter:  return 0166; // Numpad VVOD (Enter)
	}

	return 0;
  }

  switch ( key ) {
  case Qt::Key_Escape:     return 0004; // STOP key
  case Qt::Key_Shift:      return 0105; // HP key
  case Qt::Key_AsciiTilde: return 0006; // AP2 key
  case Qt::Key_Tab:	   return 0026; // TAB key
  case Qt::Key_F1:	   return 0010; // K1 / K6
  case Qt::Key_F2:	   return 0011; // K2 / K7
  case Qt::Key_F3:	   return 0012; // K3 / K8
  case Qt::Key_F4:	   return 0014; // K4 / K9
  case Qt::Key_F5:	   return 0015; // K5 / K10
  case Qt::Key_F6:         return 0152; // POM key
  case Qt::Key_F7:         return 0151; // ISP key
  case Qt::Key_F11:        return 0171; // SBROS (RESET) key
  case Qt::Key_Control:    return 0046; // SU (UPR) key
  case Qt::Key_Space:      return 0113;
  case Qt::Key_Backspace:  return 0132; // ZB (BACKSPACE) key
  case Qt::Key_Return:     return 0153; // VVOD (ENTER)
  case Qt::Key_Down:       return 0134;
  case Qt::Key_Up:         return 0154;
  case Qt::Key_Left:       return 0116;
  case Qt::Key_Right:      return 0133;
  case Qt::Key_Slash:      return 0117;
  case Qt::Key_1:          return 0030; // 1 / !
  case Qt::Key_2:          return 0031; // 2 / "
  case Qt::Key_3:          return 0032; // 3 / #
  case Qt::Key_4:          return 0013; // 4 / turtle
  case Qt::Key_5:          return 0034; // 5 / %
  case Qt::Key_6:          return 0035; // 6 / &
  case Qt::Key_7:          return 0016; // 7 / '
  case Qt::Key_8:          return 0017; // 8 / (
  case Qt::Key_9:          return 0177; // 9 / )
  case Qt::Key_0:          return 0176; // 0

  case Qt::Key_J: return 0027;
  case Qt::Key_C: return 0050;
  case Qt::Key_U: return 0051;
  case Qt::Key_K: return 0052;
  case Qt::Key_E: return 0033;
  case Qt::Key_N: return 0054;
  case Qt::Key_G: return 0055;
  case Qt::Key_BracketLeft:  return 0036;
  case Qt::Key_BracketRight: return 0037;
  case Qt::Key_Z: return 0157;
  case Qt::Key_H: return 0156;

  case Qt::Key_F: return 0047;
  case Qt::Key_Y: return 0070;
  case Qt::Key_W: return 0071;
  case Qt::Key_A: return 0072;
  case Qt::Key_P: return 0053;
  case Qt::Key_R: return 0074;
  case Qt::Key_O: return 0075;
  case Qt::Key_L: return 0056;
  case Qt::Key_D: return 0057;
  case Qt::Key_V: return 0137;

  case Qt::Key_Q: return 0067;
  case Qt::Key_S: return 0111;
  case Qt::Key_M: return 0112;
  case Qt::Key_I: return 0073;
  case Qt::Key_T: return 0114;
  case Qt::Key_X: return 0115;
  case Qt::Key_B: return 0076;
  }

  return 0;
}

void ScreenView::keyPressEvent( QKeyEvent *ev ) {
  BYTE ukncscan = translateKeyToUkncScan( ev->key(), ev->modifiers() );
  if ( ukncscan ) putKeyEventToQueue( ukncscan | (128 << 8) );
}

void ScreenView::keyReleaseEvent( QKeyEvent *ev ) {
  BYTE ukncscan = translateKeyToUkncScan( ev->key(), ev->modifiers() );
  if ( ukncscan ) putKeyEventToQueue( ukncscan );
}

void ScreenView::scanKeyboard() {
    // Process next event in the keyboard queue
    WORD keyevent = getKeyEventFromQueue();
    if (keyevent != 0) 
    {
        BOOL pressed = ((keyevent & 0x8000) != 0);
        BYTE ukncscan = LOBYTE(keyevent);
        g_pBoard->KeyboardEvent(ukncscan, pressed);
    }
}

void ScreenView::saveScreenshot( QString sFileName ) {
  img.mirrored().scaled( size() ).save( sFileName );
}
